/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <string_view>
#include <optional>
#include <utility>
#include <mutex>
#include <cstdint>
#include <memory>

namespace duck {
	class EnvBase;

	namespace detail {
		template <typename S>
		class EnvVarProxy {
		public:
			template <typename S1>
			EnvVarProxy (S1&& name, EnvBase* env) noexcept;
			EnvVarProxy& operator= (std::string_view new_val);
			operator std::string_view() const noexcept;

		private:
			S m_name;
			EnvBase* m_env;
		};
	} //namespace detail

	class EnvBase {
	public:
		typedef std::string_view mapped_type;
		typedef std::string_view key_type;
		typedef std::pair<const key_type, mapped_type> value_type;
		typedef std::size_t size_type;
		typedef std::ptrdiff_t difference_type;

		struct Mutex {
			virtual ~Mutex() noexcept = default;
			virtual void lock() = 0;
			virtual void try_lock() = 0;
			virtual void unlock() noexcept = 0;
		};

		EnvBase() = default;
		EnvBase (const EnvBase&) = delete;
		EnvBase (EnvBase&&) = delete;
		virtual ~EnvBase() noexcept = default;

		std::string_view operator[] (std::string_view name) const noexcept;
		detail::EnvVarProxy<std::string_view> operator[] (std::string_view name) noexcept;
		detail::EnvVarProxy<std::string> operator[] (std::string&& name) noexcept;
		detail::EnvVarProxy<std::string> operator[] (const std::string& name) noexcept;

		std::optional<std::string_view> get (std::string_view name) const noexcept;
		std::string_view get (std::string_view name, std::string_view def) const noexcept;
		void set (std::string_view name, std::string_view value);
		void unset (std::string_view name) noexcept;
		bool is_set (std::string_view name) const noexcept;
		size_type size() const noexcept;
		bool empty() const noexcept;
		void clear() noexcept;
		std::unique_ptr<char*[]> to_c_array() const;

		virtual std::mutex& mutex() const noexcept = 0;

	protected:
		virtual void set_all_into_implem (EnvBase& other) const = 0;
		static void set_all_into (const EnvBase& obj, EnvBase& other) { obj.set_all_into(other); }
		virtual void clear_implem(Mutex& m) = 0;
		static void clear_nolock (EnvBase& obj) noexcept;

		virtual void set_implem (std::string_view name, std::string_view value, Mutex& m) = 0;
		static void set_nolock (EnvBase& obj, std::string_view name, std::string_view value);

		virtual size_type size_implem(Mutex& m) const noexcept = 0;
		static size_type size_nolock(const EnvBase& obj) noexcept;

		virtual void unset_implem (std::string_view name, Mutex& m) = 0;
		virtual bool is_set_implem (std::string_view name, Mutex& m) const = 0;
		virtual std::optional<std::string_view> get_implem (std::string_view name, Mutex& m) const = 0;
		virtual std::string_view get_implem (std::string_view name, std::string_view def, Mutex& m) const = 0;
		virtual std::unique_ptr<char*[]> to_c_array_implem (Mutex& m) const = 0;

		virtual bool is_global() const noexcept = 0;
		virtual uint32_t class_identifier() const noexcept = 0;

	private:
		virtual void set_all_into (EnvBase& other) const;
	};

	namespace detail {
		template <typename S>
		template <typename S1>
		inline EnvVarProxy<S>::EnvVarProxy (S1&& name, EnvBase* env) noexcept :
			m_name(std::forward<S1>(name)),
			m_env(env)
		{
		}

		template <typename S>
		inline auto EnvVarProxy<S>::operator= (std::string_view new_val) -> EnvVarProxy& {
			m_env->set(m_name, new_val);
			return *this;
		}

		template <typename S>
		inline EnvVarProxy<S>::operator std::string_view() const noexcept {
			return (*m_env)[m_name];
		}
	} //namespace detail
} //namespace duck
