/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gentoofunctions.hpp"
#include "string_view_cat.hpp"
#include "config.h"
#include "run_cmd.hpp"
#include "run_context.hpp"
#include <iostream>
#if __cplusplus == 201703L
#	include <experimental/array>
#else
#	error "If make_array() has been moved out of experimental please fix"
#endif
#include <algorithm>

namespace duck {
	namespace {
		std::string_view g_last_e_cmd;

		enum YesNoIdk {
			No, Yes, Idk
		};

		std::string lowercased (std::string_view in) {
			std::string retval('\0', in.size());
			std::transform(
				in.begin(),
				in.end(),
				retval.begin(),
				[](char c)->char{return std::tolower(c);}
			);
			return retval;
		}

		template <typename I, typename T>
		bool has_item (I begin, const I& end, const T& val) {
			auto it = std::find(begin, end, val);
			return it != end;
		}

		YesNoIdk yesno_impl (std::string_view val) {
			using std::experimental::make_array;
			using std::string_view;

			auto lcval = lowercased(val);
			auto yesses = make_array<string_view>("yes", "true", "on", "1");
			auto noes = make_array<string_view>("no", "false", "off", "0");
			if (has_item(yesses.begin(), yesses.end(), lcval)) {
				return Yes;
			}
			else if (has_item(noes.begin(), noes.end(), lcval)) {
				return No;
			}
			else {
				return Idk;
			}
		}

		void generic_eprint (
			const RunContext& rc,
			std::ostream& stream,
			std::string_view quiet_var,
			std::string_view msg_prefix_var,
			std::string_view log_pri,
			std::string_view log_tag,
			std::string_view new_e_cmd,
			std::string_view msg,
			const char* newline
		) {
			if (yesno(rc, rc.env()[quiet_var])) {
				return;
			}
			else {
				const auto& env = rc.env();
				if (not yesno(rc, env["RC_ENDCOL"]) and g_last_e_cmd == "ebegin") {
					stream << '\n';
				}
				stream << ' ' << env[msg_prefix_var] << '*'
					<< env["NORMAL"] << ' ' << env["RC_INDENTATION"]
					<< msg << newline;
			}

			esyslog(rc, log_pri, log_tag, msg);
			g_last_e_cmd = new_e_cmd;
		}
	} //unnamed namespace

	void vewarn (const RunContext& rc, std::string_view msg) {
		if (yesno(rc, rc.env()["EINFO_VERBOSE"], true))
			ewarn(rc, msg);
	}

	void ewarn (const RunContext& rc, std::string_view msg) {
		generic_eprint(
			rc,
			std::clog,
			"EINFO_QUIET",
			"WARN",
			"daemon.warning",
			PROJECT_NAME,
			"ewarn",
			msg,
			"\n"
		);
	}

	void esyslog(const RunContext& rc, std::string_view pri, std::string_view tag, std::string_view msg) {
		if (not rc.env()["EINFO_LOG"].empty() and not run_cmd_retcode("command -v logger > /dev/null 2>&1")) {
			if (msg.empty())
				return;

			run_cmd("logger", "-p", pri, "-t", tag, "--", msg);
		}
	}

	bool yesno (const RunContext& rc, std::string_view var, bool force_quiet) {
		{
			const auto ret = yesno_impl(var);
			switch (ret) {
			case Yes: return true;
			case No: return false;
			default: break;
			}
		}

		{
			auto val = rc.env()[var];
			if (val.empty())
				return false;

			const auto ret = yesno_impl(val);
			switch (ret) {
			case Yes: return true;
			case No: return false;
			default:
				vewarn(rc, std::string("\"") + var + "\" (\"" + val + "\") is not set properly");
				return false;
			}
		}
	}

	void eerror (const RunContext& rc, std::string_view msg) {
		generic_eprint(
			rc,
			std::clog,
			"EERROR_QUIET",
			"BAD",
			"daemon.err",
			"rc-scripts",
			"eerrorn",
			msg,
			"\n"
		);
	}
} //namespace duck
