/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "run_cmd.hpp"
#include <cstdio>
#include <cstdlib>
#if !defined(NDEBUG)
#	include <iostream>
#	include <algorithm>
#endif
#include <stdexcept>
#include <memory>
#include <pstream.h>
#include <array>
#include <cassert>

namespace duck::detail {
	namespace {
		void trim_eol (std::string& val) {
			const auto pos = val.find_last_not_of('\n');
			if (val.npos != pos)
				val.resize(pos + 1);
		}
	} //unnamed namespace

	RunCmdOutput run_cmd(const std::string& cmd, const std::vector<std::string>& argv) {
		assert(not argv.empty());
		assert(argv.front() == cmd);

#if !defined(NDEBUG)
		std::cout << "Running command: " << cmd;
		std::for_each(argv.begin(), argv.end(), [](const auto& s){std::cout << ' ' << s;});
		std::cout << '\n';
#endif

		const redi::pstream::pmode mode = redi::pstreams::pstdout | redi::pstreams::pstderr;
		redi::ipstream ips(cmd, argv, mode);

		std::streamsize n;
		RunCmdOutput ret;
		std::array<char, 64> buf;
		std::array<bool, 2> finished {false, false};

		while (not finished[0] or not finished[1]) {
			if (not finished[0]) {
				while ((n = ips.err().readsome(buf.data(), buf.size())) > 0)
					ret.err.append(buf.begin(), buf.begin() + n);
				if (ips.eof()) {
					finished[0] = true;
					if (not finished[1])
						ips.clear();
				}
			}

			if (!finished[1]) {
				while ((n = ips.out().readsome(buf.data(), buf.size())) > 0)
					ret.out.append(buf.begin(), buf.begin() + n);
				if (ips.eof()) {
					finished[1] = true;
					if (!finished[0])
						ips.clear();
				}
			}
		}

		trim_eol(ret.out);
		trim_eol(ret.err);
		return ret;
	}
} //namespace detail::duck

namespace duck {
	int run_cmd_retcode (const char* command) {
		assert(command);
		return std::system(command);
	}
} //namespace duck
