/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "run_context.hpp"
#include "to_var_map.hpp"
#include "string_view_cat.hpp"
#include <unistd.h>
#include <cassert>

namespace duck {
	namespace {
		std::string fix_eprefix(std::string_view p) {
			if (p.size() >= 2 and p.front() == '@' and p.back() == '@')
				return "";
			else
				return std::string(p);
		}

		std::string_view remove_front_slash(std::string_view text) {
			const std::size_t pos = text.find_first_not_of('/');
			return (text.npos == pos ? text : text.substr(pos));
		}
	} //unnamed namespace

	RunContext::RunContext() :
		m_env(),
		m_root(m_env.get("ROOT", "/")),
		m_eprefix(fix_eprefix(m_env.get("EPREFIX", ""))),
		m_eroot(m_root / remove_front_slash(std::string(m_eprefix))),
		m_env_d(m_eroot / "etc/env.d"),
		m_gcc_env_d(m_env_d / "gcc")
	{
	}

	void RunContext::add_to_env (std::string_view colours) {
		for (auto& entry : to_var_map_ignore_ownership(colours)) {
			m_env.set(entry.first, entry.second);
		}
	}

	auto RunContext::root() const -> const path_type& {
		return m_root;
	}

	auto RunContext::eroot() const -> const path_type& {
		return m_eroot;
	}

	auto RunContext::eprefix() const -> const path_type& {
		return m_eprefix;
	}

	auto RunContext::env_d() const -> const path_type& {
		return m_env_d;
	}

	auto RunContext::gcc_env_d() const -> const path_type& {
		return m_gcc_env_d;
	}

#if !defined(NDEBUG)
	std::ostream& operator<< (std::ostream& oss, const RunContext& sp) {
		oss << "RunContext {\n" <<
			"\tROOT: " << sp.root() << '\n' <<
			"\tEPREFIX: " << sp.eprefix() << '\n' <<
			"\tEROOT: " << sp.eroot() << '\n' <<
			"\tENV_D: " << sp.env_d() << '\n' <<
			"\tGCC_ENV_D: " << sp.gcc_env_d() << '\n' <<
			'}';
		return oss;
	}
#endif
} //namespace duck
