/* Copyright 2020, Michele Santullo
 * This file is part of user-gcc.
 *
 * User-gcc is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * User-gcc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with user-gcc.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "to_var_map.hpp"
#include "string_view_cat.hpp"
#include <stdexcept>
#include <cassert>
#include <cctype>
#include <algorithm>

namespace duck {
	namespace {
		std::string_view trimmed (std::string_view text) {
			using std::find_if_not;

			auto isspace_func = [](char c) {return std::isspace(c);};
			const auto new_start_it = find_if_not(text.cbegin(), text.cend(), isspace_func);
			const std::size_t new_start = (new_start_it == text.cend() ? text.size() : static_cast<std::size_t>(new_start_it - text.cbegin()));

			const auto new_end_it = find_if_not(text.crbegin(), text.crend(), isspace_func);
			const std::size_t new_end = text.size() - (new_end_it == text.crend() ? std::size_t{} : static_cast<std::size_t>(new_end_it - text.crbegin()));

			assert(new_end <= text.size());
			assert(new_end >= new_start);
			return text.substr(new_start, new_end - new_start);
		}

		std::pair<std::string_view, std::string_view> split_key_val(std::string_view text) {
			const auto eq_pos = text.find('=');
			if (eq_pos == text.npos) {
				throw std::runtime_error("Unrecognised line \"" + text + "\": expected key=value format not found");
			}

			return std::make_pair(
				text.substr(0, eq_pos),
				text.substr(eq_pos + 1)
			);
		}

		std::string_view unquoted (std::string_view text) {
			std::string_view ret = trimmed(text);
			if (ret.size() < 2)
				return ret;
			else if (ret.front() == '"' and ret.back() == '"')
				return ret.substr(1, ret.size() - 2);
			else
				return ret;
		}

		void add_key_val_text_to_map (std::string_view text, VarMap& out) {
			std::string_view line = trimmed(text);
			if (line.empty() or line.front() == '#')
				return;

			auto new_item = split_key_val(line);
			out[new_item.first] = unquoted(new_item.second);
		}
	} //unnamed namespace

	VarMap to_var_map (std::string& text) {
		return to_var_map_ignore_ownership(text);
	}

	VarMap to_var_map_ignore_ownership (std::string_view text) {
		VarMap ret;
		std::size_t pos = 0;
		std::size_t start = 0;
		while ((pos = text.find('\n', pos)) != text.npos) {
			add_key_val_text_to_map(text.substr(start, pos - start), ret);
			start = pos + 1;
			++pos;
		}

		add_key_val_text_to_map(text.substr(start), ret);
		return ret;
	}

	VarMap to_var_map_ignore_ownership (char** env) {
		assert(env);

		VarMap ret;
		std::size_t z = 0;
		while (env[z]) {
			std::string_view line(env[z]);
			add_key_val_text_to_map(line, ret);
			++z;
		}
		return ret;
	}
} //namespace duck
